import numpy as np
import torch
import random
from torch.autograd import Variable
import os
from gan import Generator, Discriminator
from utils import load_data, save_samples_midi, write_to_file, plot_results
from train import run_epoch

# System configurations
config = {
        "learning_rate": 0.001,
        "hidden_dim": 512,
        "dropout_prob": 0.5,
        "epochs": 50,
        "freeze": True,
        "feature_matching": True,
        "label_smoothing": True,
        "pretraining": True,
        "pretraining_epochs": 6,
        "time_steps": 48,
        "num_samples": 125,
        "test_size": 0.5,
        "gen_samples": 5,
        "reduce_dim": False,
        "prev_timesteps": False,
        "device": "cuda" if torch.cuda.is_available() else "cpu",
        "data_path": "../data/training_data.npy"

    }

def main(data, config):
    # Reduced tonal range or not
    tone_range = 24 if config["reduce_dim"] else 84

    model = {
        "g": Generator(tone_range*config["time_steps"], config["hidden_dim"], config["prev_timesteps"], config["device"], tone_range, config["time_steps"], config["dropout_prob"], ).to(config["device"]),
        "d": Discriminator(config["hidden_dim"], tone_range*config["time_steps"], config["device"], config["time_steps"], tone_range, config["dropout_prob"]).to(config["device"])
    }

    
    optimizer = {
        "g": torch.optim.Adam(model['g'].parameters(), lr=config["learning_rate"]),
        "d": torch.optim.Adam(model['d'].parameters(), lr=config["learning_rate"])
    } 

    criterion = torch.nn.BCEWithLogitsLoss().to(config["device"])

    Tensor = torch.cuda.FloatTensor if torch.cuda.is_available() else torch.FloatTensor

    # Label smoothing technique
    if config["label_smoothing"]:
        truths = Variable(Tensor(384, 1, 1, 1).fill_(0.9), requires_grad=False)
        fakes = Variable(Tensor(384, 1, 1, 1).fill_(0.0), requires_grad=False)

    else:
        truths = Variable(Tensor(384, 1, 1, 1).fill_(1.0), requires_grad=False)
        fakes = Variable(Tensor(384, 1, 1, 1).fill_(0.0), requires_grad=False)

    data['truths'] = truths
    data['fakes'] = fakes

    d_loss_arr = []
    g_loss_arr = []

    print("Starting training on {}...".format(config["device"]))

    total_corr_arr = []
    total_acc_arr = []
    
    pretrain = {
        'g_loss_fg': [],
        'd_loss_fg': [],
        'd_acc_fg': [],
        'd_corr_fg': [],
        'g_loss_fd': [],
        'd_loss_fd': [],
        'd_acc_fd': [],
        'd_corr_fd':[]
    }

    # Add losses and accuracy for the pretraining epochs
    if config["pretraining"]:
        for epoch in range(config["pretraining_epochs"]):
            model, pretrain_loss_fg, pretrain_corr_fg, pretrain_val_loss_fg, pretrain_val_corr_fg, = run_epoch(model, optimizer, criterion, data, epoch, config["pretraining_epochs"], freeze_g=True, pretraining=True, feature_matching=config["feature_matching"])
            pretrain['g_loss_fg'].append(pretrain_loss_fg['g'])
            pretrain['d_loss_fg'].append(pretrain_loss_fg['d'])
            pretrain['d_acc_fg'].append(pretrain_corr_fg['accuracy'])
            pretrain['d_corr_fg'].append(pretrain_corr_fg['num_corrects'])

        for epoch in range(config["pretraining_epochs"]):
            model, pretrain_loss_fd, pretrain_corr_fd, pretrain_val_loss_fd, pretrain_val_corr_fd = run_epoch(model, optimizer, criterion, data, epoch, config["pretraining_epochs"], freeze_d=True, pretraining=True)
            pretrain['g_loss_fd'].append(pretrain_loss_fd['g'])
            pretrain['d_loss_fd'].append(pretrain_loss_fd['d'])
            pretrain['d_acc_fd'].append(pretrain_corr_fd['accuracy'])
            pretrain['d_corr_fd'].append(pretrain_corr_fd['num_corrects'])
    
    freeze_d = False

    for epoch in range(config["epochs"]):

        model, loss, corrects, val_loss, val_corr = run_epoch(model, optimizer, criterion, data, epoch, config["epochs"], freeze_d=freeze_d, feature_matching=config["feature_matching"])
        if config["freeze"]:
            freeze_d = False
            # Discriminator is frozen if accuracy > 95 %
            if corrects["accuracy"] >= 95.0:
                freeze_d = True
        
        d_loss_arr.append(loss['d'])
        g_loss_arr.append(loss['g'])

        total_corr_arr.append(corrects["num_corrects"])
        total_acc_arr.append(corrects["accuracy"])
        
    return model, d_loss_arr, g_loss_arr, total_corr_arr, total_acc_arr, pretrain

if __name__ == "__main__":
    data = load_data(config["data_path"], config["num_samples"], config["device"], test_size = config["test_size"], reduce_dim = config["reduce_dim"])

    model, d_loss_arr, g_loss_arr, total_corr_arr, total_acc_arr, pretrain = main(data, config)
    # Generate sample accompaniments
    save_samples_midi(model['g'], data, config)

    # Save losses and accuracy
    write_to_file(g_loss_arr, "g_loss", config)
    write_to_file(d_loss_arr, "d_loss", config)
    write_to_file(total_corr_arr, "corrects", config)
    write_to_file(total_acc_arr, "accuracy", config)
    
    # Save losses and accuracy from pretraining
    write_to_file(pretrain['g_loss_fg'], "g_loss_freeze_g", config)
    write_to_file(pretrain['d_loss_fg'], "d_loss_freeze_g", config)
    write_to_file(pretrain['d_acc_fg'], "accuracy_freeze_g", config)
    write_to_file(pretrain['d_corr_fg'], "corrects_freeze_g", config)
    
    write_to_file(pretrain['g_loss_fd'], "g_loss_freeze_d", config)
    write_to_file(pretrain['d_loss_fd'], "d_loss_freeze_d", config)
    write_to_file(pretrain['d_acc_fd'], "accuracy_freeze_d", config)
    write_to_file(pretrain['d_corr_fd'], "corrects_freeze_d", config)
    
    # Plot the losses and accuracy
    plot_results(g_loss_arr, d_loss_arr, total_corr_arr, total_acc_arr, pretrain, config)