
import numpy as np
import torch
from torch import optim
import torch.nn as nn
from torch.autograd import Variable

# Maximum gradient
MAX_GRAD_NORM = 5.0


def run_training(model, optimizer, criterion, data, freeze_g=False, freeze_d=False, feature_matching=False):

    g_loss_total = 0.
    d_loss_total = 0.
    
    loss = {
        'g': 0.,
        'd': 0.
    }
    
    corr_total = 0.
    corr_acc_total = 0.
    for i in range(len(data['train_x'])):
        model['g'].train()
        model['d'].train()

        state_G = model['g'].init_hidden()
        state_D = model['d'].init_hidden()

        if not freeze_g:
            optimizer['g'].zero_grad()
        generated, state_G = model['g'](data['train_x'][i], state_G)
        generated = torch.squeeze(generated)

        dis_fake_output, dis_fake,_ = model['d'](generated.detach(), state_D)
        dis_real_output, dis_real, _ = model['d'](data['train_y'][i], state_D)

        if feature_matching:
            g_loss = criterion(dis_real, dis_fake)
        else:
            g_loss = criterion(dis_fake_output, data['truths'])

        if not freeze_g:
            g_loss.backward(retain_graph=True)
            nn.utils.clip_grad_norm_(model['g'].parameters(), max_norm=MAX_GRAD_NORM)
            optimizer['g'].step()
        if not freeze_d:
            optimizer['d'].zero_grad()
        
        real_loss = criterion(dis_real_output, data['truths'])
        fake_loss = criterion(dis_fake_output, data['fakes'])
        
        d_loss = (real_loss + fake_loss) / 2
        if not freeze_d:
            d_loss.backward()
            nn.utils.clip_grad_norm_(model['d'].parameters(), max_norm=MAX_GRAD_NORM)
            optimizer['d'].step()
        
        d_loss_total += d_loss.item()
        g_loss_total += g_loss.item()
        
        # The number of correctly labeled by the discriminator
        num_corrects = (dis_real_output > 0.5).sum().item() + (dis_fake_output < 0.5).sum().item()
        
        d_acc = 100 * num_corrects / (2 * 384)
        
        corr_acc_total += d_acc
        corr_total += num_corrects
    
    avg_acc = corr_acc_total / len(data['train_x'])
    avg_corr = corr_total / len(data['train_x'])
    
    corr_dict = {
        "accuracy": avg_acc,
        "num_corrects": avg_corr
    }
    
    loss['d'] = d_loss_total / len(data['train_x'])
    loss['g'] = g_loss_total / len(data['train_x'])
    
    return model, loss, corr_dict
    
def run_validation(model, criterion, data, feature_matching=False):
    model['g'].eval()
    model['d'].eval()
    
    corr_dict = {
        "accuracy": 0.,
        "num_corrects": 0.
    }
    
    g_loss_total = 0.
    d_loss_total = 0.
    

    corr_total = 0.
    corr_acc_total = 0.
    
    for i in range(len(data['test_x'])):
        state_G = model['g'].init_hidden()
        state_D = model['d'].init_hidden()
        
        generated, _ = model['g'](data['test_x'][i], state_G)
        
        dis_real_output, dis_real,_ = model['d'](data['test_y'][i], state_D)
        dis_fake_output, dis_fake,_ = model['d'](generated, state_D)
        
        if feature_matching:
            g_loss = criterion(dis_fake, dis_real)
        else:
            g_loss = criterion(dis_fake_output, data['truths'])
        
        real_loss = criterion(dis_real_output, data['truths'])
        fake_loss = criterion(dis_fake_output, data['fakes'])
        
        d_loss = (real_loss + fake_loss) / 2
        g_loss_total += g_loss.item()
        d_loss_total += d_loss.item()
        
        # How many are correctly labeled by the discriminator
        num_correct = (dis_real_output > 0.5).sum().item() + (dis_fake_output < 0.5).sum().item()
        d_acc = 100 * num_correct / (2 * 384)
        
        corr_acc_total += d_acc
        corr_total += num_correct
        
    avg_acc = corr_acc_total / len(data['test_x'])
    avg_corr = corr_total / len(data['test_x'])
    
    corr_dict = {
        "accuracy": avg_acc,
        "num_corrects": avg_corr
    }

    d_loss_avg = d_loss_total / len(data['test_x'])
    g_loss_avg = g_loss_total / len(data['test_x'])
    
    loss = {
        'g': g_loss_avg,
        'd': d_loss_avg
    }
    return loss, corr_dict

def run_epoch(model, optimizer, criterion, data, epoch, num_epochs, freeze_g = False, freeze_d = False, pretraining=False, feature_matching=False):
    
    model, loss, corrects = run_training(model, optimizer, criterion, data, freeze_g=freeze_g, freeze_d = freeze_d, feature_matching=feature_matching)
    val_loss, val_corr = run_validation(model, criterion, data, feature_matching = feature_matching)
    
    if pretraining:
        print("PRETRAINING EPOCH {0} / {1}:\tFreeze G: {2}, Freeze D: {3}".format(epoch+1, num_epochs, freeze_g, freeze_d))
    else:
        print("TRAINING EPOCH {0} / {1}:\tFreeze G: {2}, Freeze D: {3}".format(epoch+1, num_epochs, freeze_g, freeze_d))
    print("Training:\tG loss: {0}, D loss: {1}, D accuracy {2}:".format(loss['g'], loss['d'], corrects['accuracy']))
    print("Validation:\tG loss: {0}, D loss: {1}, D accuracy {2}:".format(val_loss['g'], val_loss['d'], val_corr['accuracy']))
    print("-----------------------------------------------------")
    
    return model, loss, corrects, val_loss, val_corr
